from typing import Optional

from mag_tools.log.logger import Logger
from mag_tools.model.log_type import LogType
from mag_tools.utils.common.time_probe import TimeProbe
from selenium.common.exceptions import InvalidSelectorException, NoSuchElementException, TimeoutException, \
    WebDriverException
from selenium.webdriver import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement

from mag_test.bean.control import Control
from mag_test.bean.element_info import ElementInfo
from mag_test.core.app_driver import AppDriver
from mag_test.finder.driver_finder_utils import DriverFinderUtils
from mag_test.finder.element_finder_utils import ElementFinderUtils
from mag_test.model.control_type import ControlType
from mag_test.utils.event_utils import EventUtils
from mag_test.utils.tree_utils import TreeUtils
from mag_test.model.init_status import InitStatus
from mag_test.model.menu_type import MenuType
from mag_test.utils.element_utils import ElementUtils


class ElementFinder:
    @staticmethod
    def find(driver:AppDriver, element_info: ElementInfo):
        Logger.debug(f'开始查找控件：{element_info.main_info}')
        probe = TimeProbe.get_probe('查找控件')

        # 菜单
        if element_info.main_type == ControlType.MENU:
            ElementFinder.__find_menu(driver, element_info)
        # 窗口
        elif element_info.main_type == ControlType.WINDOW:
            ElementFinder.__find_window(driver, element_info)
        # 日期时间
        elif element_info.main_type == ControlType.DATETIME:
            ElementFinder.__find_datetime(driver, element_info)
        else:
            # 查找父控件并处理响应事件
            parent = ElementFinder.__find_parent(driver, element_info.parent_info)

            # 查找主控件并处理响应事件
            main_element = ElementFinder.__find_main_element(driver, parent, element_info.main_info, element_info.value)

            # 查找子控件并处理响应事件
            child_element = ElementFinder.__find_child_element(driver, main_element, element_info.child_info, element_info.value)

            # 查找弹出菜单并处理响应事件
            ElementFinder.__find_context_menu(driver, child_element, element_info.pop_menu)
        probe.write_log()

    @staticmethod
    def __find_menu(driver:AppDriver, element_info:ElementInfo):
        """
        查找菜单控件
        参数：
        name 菜单及菜单项名，格式：菜单项名/子菜单项名/...
        parent_name 父控件名
        parent_control_type 父控件类型
        """
        items = element_info.menu_items

        element = None
        actions = ActionChains(driver)

        for index, item in enumerate(items):
            element = DriverFinderUtils.find_element_by_type_wait(driver, item, ControlType.MENU_ITEM)
            if index < len(items) - 1:
                actions.move_to_element(element).click().perform()
        return element

    @staticmethod
    def __find_datetime(driver:AppDriver, element_info:ElementInfo):
        """
        查找日期时间控件
        参数：
        name 日期时间控件名
        """
        parent = ElementFinder.__find_parent(driver, element_info.parent_info)
        if parent:
            dt = ElementFinderUtils.find_element_by_class(parent, element_info.main_name, 'SysDateTimePick32')
        else:
            dt = DriverFinderUtils.find_element_by_class(driver, element_info.main_name, 'SysDateTimePick32')

        return dt

    @staticmethod
    def __find_window(driver:AppDriver, element_info: ElementInfo):
        """
        查找窗口
        参数：
        name 窗口名（关键词）
        """
        return driver.find_element(By.XPATH, f"//Window[contains(@Name, '{element_info.main_name}')]")
    
    @staticmethod
    def __find_context_menu(driver:AppDriver, element: WebElement, menu_Info: Optional[Control]):
        if menu_Info and menu_Info.menu_type:
            actions = ActionChains(driver)
            actions.move_to_element(element).context_click(element).perform()

            try:
                menu = None
                if menu_Info.menu_type == MenuType.CONTEXT:
                    menu = DriverFinderUtils.find_element_by_class(driver.root_driver, '上下文', '#32768')
                elif menu_Info.menu_type == MenuType.POPUP:
                    menu = DriverFinderUtils.find_element_by_type(driver, '弹出窗口', ControlType.WINDOW)

                if menu:
                    if menu_Info.automation_id:
                        menu_item = ElementFinderUtils.find_element_by_automation(menu, menu_Info.automation_id)
                    else:
                        menu_item = ElementFinderUtils.find_element_by_type(menu, menu_Info.name, ControlType.MENU_ITEM)

                    Logger.debug(LogType.FRAME, f"控件的菜单类型为：{menu_Info.menu_type.desc}")
                    menu_item.click()
            except (NoSuchElementException, InvalidSelectorException, WebDriverException) as e :
                Logger.throw(f"未找到菜单：{str(e)}")

    @staticmethod
    def __find_child_element(driver: AppDriver, main_element: WebElement, child_info: Control, value: Optional[str] = None):
        """
        查找树视图
        参数：
        name 树视图名，格式：树视图名/树节点名/菜单名，树视图名和菜单名可为空，只支持菜单模式的弹出菜单
        """
        if child_info:
            if child_info.action == InitStatus.EXPANDED:
                TreeUtils.expand_all(driver, main_element)

            if child_info.automation_id:
                element = ElementFinderUtils.find_element_by_automation(main_element, child_info.automation_id)
            else:
                element = ElementFinderUtils.find_element_by_type(main_element, child_info.name, child_info.control_type)

            EventUtils.process_final_element(driver, element, child_info, value)
        else:
            element = main_element

        return element

    @staticmethod
    def __find_main_element(driver: AppDriver, parent: WebElement, main_info: Control, value: Optional[str] = None) -> WebElement:
        try:
            if main_info.automation_id:
                if parent:
                    element = ElementFinderUtils.find_element_by_automation(parent, main_info.automation_id)
                else:
                    element = DriverFinderUtils.find_element_by_automation_wait(driver, main_info.automation_id)
            else:
                if parent:
                    element = ElementFinderUtils.find_element_by_type(parent, main_info.name, main_info.control_type)
                else:
                    element = DriverFinderUtils.find_element_by_type_wait(driver, main_info.name, main_info.control_type)

            if main_info.is_composite:
                offset = ElementUtils.get_main_offset(value, element)
                EventUtils.click_offset(driver, element, offset)
            else:
                EventUtils.process_final_element(driver, element, main_info, value)

            return element
        except NoSuchElementException as e:
            Logger.throw(f'未找到主控件：{str(e)}')    # 模糊查找控件时，返回None正常
        except InvalidSelectorException as e:
            Logger.throw(f'无效的控件选项：{str(e)}')  # 模糊查找控件时，返回None正常
        except TimeoutException as e:
            Logger.throw(f'连接失败：{str(e)}')  # 模糊查找控件时，返回None正常
        except WebDriverException as e:
            Logger.throw(f'连接失败或超时：{str(e)}')   # 模糊查找控件时，返回None正常

    @staticmethod
    def __find_parent(driver:AppDriver, parent_info: Control):
        parent = None
        if parent_info:
            try:
                Logger.debug(f'开始查找父控件：{parent_info}')
                probe = TimeProbe.get_probe('查找父控件')

                if parent_info.automation_id:
                    parent = DriverFinderUtils.find_element_by_automation_wait(driver, parent_info.automation_id)
                elif parent_info.control_type:
                    parent = DriverFinderUtils.find_element_by_type_wait(driver, parent_info.name, parent_info.control_type)

                probe.write_log()
            except (NoSuchElementException, InvalidSelectorException, WebDriverException) as e:
                Logger.throw(f'查找父控件失败：{parent_info}\n{str(e)}')

        return parent


import os

import pandas as pd
from mag_tools.log.logger import Logger
from mag_tools.model.log_type import LogType
from mag_tools.utils.common.string_utils import StringUtils
from mag_tools.utils.file.excel_utils import ExcelUtils
from mag_tools.utils.file.json_file_utils import JsonFileUtils

from mag_test.model.control_type import ControlType
from mag_test.model.usage_status import UsageStatus


class ExcelConvertor:
    def __init__(self, home_dir: str):
        self.__home_dir = home_dir

    def convert(self, plan_excel_file:str, test_type:str, sheet_name:str):
        excel_file = self.__get_excel_file(plan_excel_file)

        plan_id = self.__convert_plan(excel_file, test_type, sheet_name)
        self.__convert_function(excel_file, plan_id)

        Logger.info(LogType.FRAME, "将Excel格式的测试计划转换为JSon脚本完毕", True)

        return plan_id

    def __convert_plan(self, excel_file:str, test_type:str, sheet_name:str):
        df = pd.read_excel(excel_file, sheet_name)  # 读取 Excel 文件
        df = ExcelUtils.clear_empty(df) # 清除空白的行与列

        title = ExcelUtils.get_first_cell_of_row_by_keyword(df, 0, sheet_name)
        plan_name, plan_id = StringUtils.split_name_id(title)
        if not plan_id:
            plan_id = StringUtils.pick_head(plan_name, sheet_name) if test_type.upper() == "ALL" else "unit"

        df.columns = df.iloc[1]  # 将第1行设置为列名
        df = df.drop(index=range(2)).reset_index(drop=True) # 删除第1行及其之前的所有行

        # 按模块分组
        df["功能模块"] = df["功能模块"].ffill()
        df["测试用例"] = df["测试用例"].ffill()
        df["备注"] = df["备注"].ffill()

        if "用例状态" in df.columns:
            df["用例状态"] = df["用例状态"].ffill()

        module_groups = df.groupby("功能模块", sort=False)

        _modules = []
        for _module_name, module_group in module_groups:
            Logger.debug(LogType.FRAME, f"功能模块名：{_module_name}")

            _cases = []
            case_groups = module_group.groupby("测试用例", sort=False)
            for _case_name, case_group in case_groups:
                Logger.debug(LogType.FRAME, f"测试用例名：{_case_name}")

                _case_desc = ExcelUtils.get_value_from_group(case_group, "备注")
                _case_status = UsageStatus.of_desc(ExcelUtils.get_value_from_group(case_group, "用例状态", "正常"))

                _functions = []
                function_groups = case_group.groupby("功能", sort=False)
                for _function_name, function_group in function_groups:
                    _function_name, _function_id = StringUtils.split_name_id(str(_function_name))
                    # 功能标识为空时忽略
                    if _function_id:
                        _functions.append({"id": _function_id, "name": _function_name})

                _case = {"name": _case_name, "desc": _case_desc, "functions": _functions, "status": _case_status}
                _cases.append(_case)
            _module = {"name": _module_name, "cases": _cases}
            _modules.append(_module)

        _plan = {"name": plan_name, "modules": _modules}

        JsonFileUtils.save_as_json(self.__get_plan_file(plan_id), _plan)

        return plan_id

    def __convert_function(self, excel_file, plan_id):
        df = pd.read_excel(excel_file, "功能")  # 读取 Excel 文件
        df = ExcelUtils.clear_empty(df) # 清除空白的行与列
        df = ExcelUtils.delete_row_with_keyword(df, 0, '测试功能')

        df.columns = df.iloc[0]  # 将第1行设置为列名
        df = df.drop(index=range(1)).reset_index(drop=True) # 删除第0行及其之前的所有行

        # 按模块分组
        df["功能"] = df["功能"].ffill()
        if "功能状态" in df.columns:
            df["功能状态"] = df["功能状态"].ffill()
        function_groups = df.groupby("功能", sort=False)

        for _function_name, function_group in function_groups:
            Logger.debug(LogType.FRAME, f"功能：{_function_name}")
            _function_status = UsageStatus.of_desc(ExcelUtils.get_value_from_group(function_group, "功能状态", "正常"))

            test_steps = []
            step_groups = function_group.groupby("步骤", sort=False)
            for step_name, step_group in step_groups:
                Logger.debug(LogType.FRAME, f"功能步骤：{step_name}")

                _element_name = ExcelUtils.get_value_from_group(step_group, "控件名")
                _element_type = ControlType.of_desc(ExcelUtils.get_value_from_group(step_group, "控件类型"))
                _element_id = ExcelUtils.get_value_from_group(step_group, "控件标识")
                _element_parent_name = ExcelUtils.get_value_from_group(step_group, "父控件名")
                _element_parent_type = ControlType.of_desc(ExcelUtils.get_value_from_group(step_group, "父控件类型"))
                _element_parent_id = ExcelUtils.get_value_from_group(step_group, "父控件标识")
                _element_value = ExcelUtils.get_value_from_group(step_group, "输入")
                _element_pop = ExcelUtils.get_value_from_group(step_group, "弹窗")
                _element_status = UsageStatus.of_desc(ExcelUtils.get_value_from_group(step_group, "步骤状态", "正常"))

                test_step = {"control_name": str(_element_name) if _element_id else None,
                            "control_type": _element_type.name if _element_type else None,
                            "id": str(_element_id) if _element_id else None,
                            "parent": str(_element_parent_name) if _element_parent_type else None,
                            "parent_type": _element_parent_type.name if _element_parent_type else None,
                            "parent_id": str(_element_parent_id) if _element_parent_id else None,
                            "value": str(_element_value) if _element_value else None,
                            "step_name": str(step_name) if _element_name else None,
                            "pop": str(_element_pop) if _element_pop else None,
                            "status": _element_status}
                test_steps.append(test_step)

            _function_name, _function_id = StringUtils.split_name_id(str(_function_name))
            _function = {"name": _function_name, "steps": test_steps, "status": _function_status}

            # 功能标识为空时忽略
            if _function_id:
                JsonFileUtils.save_as_json(self.__get_function_file(plan_id, _function_id), _function)

    def __get_excel_file(self, excel_file):
        if ":" not in excel_file:
            excel_dir = os.path.join(self.__home_dir, 'excel')
            excel_file = os.path.join(excel_dir, excel_file)
        return excel_file

    def __get_plan_file(self, plan_id):
        script_dir = os.path.join(self.__home_dir, 'script')
        return os.path.join(script_dir, plan_id, 'plan.json')

    def __get_function_file(self, plan_id, function_name):
        script_dir = os.path.join(self.__home_dir, 'script')
        return os.path.join(script_dir, plan_id, f'{function_name}.json')
